#
/*
 *	    M U L T I - D E V I C E   Q U E U E I N G   S Y S T E M
 *
 *			   Douglas P. Kingston III
 *
 *		Qume-style Printer Server Process  (QUMESERVER)
 *
 *	Note that a pipe to the Deamon will be on file descriptor 3.
 *	________________________________________________________________
 *
 *				Revision History
 *				================
 */

#include <stdio.h>
#include <pwd.h>
#include <signal.h>
#include "queue.h"
#include "qreplys.h"

time_t	timenow;
struct	request r;
int	lpoptions;
long	linecount = 0L;	/* number of lines printed by this server */
char	source[LINESIZ]; /* source of request, account name or net address */
char	obuf[BUFSIZ];	/* remember the output goes to stdout */
char	*arg0;		/* Pointer to the original argv[0] */
char	*title = 0;

#if BSD42 | BSD41c | V7
#include <sgtty.h>
struct sgttyb ttyb =	/* tty line setup commands */
{
	B1200,B1200,
	0,0,
	TANDEM|CRMOD|XTABS|ANYP		/*default parity */
};
#endif
#if BRLUNIX
#include <sgtty.h>
struct sg_brl ttyb =	/* tty line setup commands */
{
	B1200,B1200,
	DC3DC1|STALL,
	CRMOD|XTABS|ANYP
};
#endif

struct	request *gethdr();
char	*index();
char	*strdup();
int	doflush();
int	dorestart();

extern	char *Qdatadir;

#define	FF	014	/* A formfeed character */
#define CR	015	/* A return character */

/*
 *		M A I N  ( )
 *  As a MDQS queue server, this process will be started up
 *  with the queue control file as the standard input, and
 *  the device of interest on the standard output.
 */
main (argc, argv)
int	argc;
char	**argv;
{
	char	datafile[LINESIZ];
	char	linebuf[LINESIZ];
	char	msg[LINESIZ];
	char	*cp;
	
	signal( SIGHUP, SIG_IGN );		/* Paranoid. */
	signal( SIGEMT, dorestart );
	signal( SIGTERM, doflush );

	arginit(argc, argv);
	time( &timenow );
	setbuf( stdout, obuf );

	stty( fileno( stdout ), &ttyb );

	if (gethdr (stdin, &r) == (struct request *) NULL) {
		err( "%s: Error reading request", arg0);
		retmsg (RP_FATAL, "Error reading request header");
		exit (10);
	}

	while (fgets (linebuf, sizeof (linebuf), stdin) != NULL) {
		if ((cp = index (linebuf, '\n')) != 0)
			*cp = 0;
		switch (linebuf[0]) {
		case 'D':
			if( chdir( &linebuf[1] ) < 0 ) {
				printf("\n\nError trying chdir (%s,%s)\n%c",
				&linebuf[1], sys_errlist[errno], FF );
				chdir( "/tmp" );	/* A safe default */
			}
			break;

		case 'O':
			setoptions (&linebuf[1]);
			break;

		case 'I':
			sprintf (msg, "printing %s", title);
			statmsg (msg);
			sprintf (datafile, "%s/%s", Qdatadir, &linebuf[1]);
			print (datafile);
			unlink (datafile);
			break;

		case 'F':
			sprintf (msg, "printing %s", &linebuf[1]);
			statmsg (msg);
			if (print (&linebuf[1]) < 0)
				break;
			if (lpoptions & NOKEEP)
				unlink (&linebuf[1]);
			break;

		case 'T':
			if (title) free(title);
			title = strdup(&linebuf[1]);
			break;

		case 'X':
			doxarg (&linebuf[1]);
			break;

		case 'B':
		case 'H':
		case 'U':
		default:
			/* Ignore unknown lines */
			break;
		}
	}
	retmsg (RP_OK, 0);
	exit (0);
}

/*
 *		S E T O P T I O N S  ( )
 *  Setoptions is passed the ascii string numerical representation
 *  of the lpoptions flags word.  The bit definitions are in queue.h.
 *  The string is a decimal representation.
 */
setoptions (optstr)
char *optstr;
{
	lpoptions = atoi (optstr);
}

/*
 *		D O X A R G  ( )
 *  Parse and act on extended arguments.
 */
doxarg (strp)
register char *strp;
{
	if (strcmp (strp, "nocrmod")==0)
		ttyb.sg_flags &= ~CRMOD;
	stty( fileno( stdout ), &ttyb );
}

/*
 *		P R I N T  ( )
 *  Dump a file onto the printer.  We do no checking
 *  as were are likely to be sending binary printer codes.
 */
print( file )
char *file;
{
	FILE	*df;
	int	cnt;

	if ((df = fopen (file, "r")) == NULL) {
		printf ("\n\n***** Could not open input file (%s, %s) *****\n",
			file, sys_errlist[errno] );
		putchar( FF );
		return( -1 );
	}
	for( cnt = r.r_copies; cnt > 0; cnt--) {
		if (copy (df, stdout) < 0)
			return (-1);
		if( cnt ) {
			putchar( FF );
			rewind( df );
		}
	}
	fclose( df );
	putchar( FF );
	return( 0 );
}
/*
 *		C O P Y ( )
 *  Takes two file pointers, and copies from one to the other
 */
copy (from, to)
FILE *from, *to;
{
	char	buf[BUFSIZ];
	int	nread;

	while ((nread = fread (buf, 1, BUFSIZ, from)) > NULL) {
		fwrite (buf, 1, nread, to);
		if (ferror(to)) {
			err ("Error reading source file (%s)",
				sys_errlist[errno]);
			return (-1);
		}
		if (ferror(from)) {
			err ("Error writing to printer (%s)",
				sys_errlist[errno]);
			return (-1);
		}
	}
	return (0);
}

/*
 *	A R G I N I T
 *
 *	Parse args of lineprinter like servers
 */
arginit(argc, argv)
int	argc;
char	**argv;
{
	int	baud;
	char	*parp;

	arg0 = argv[0];

	while (*++argv) {
		if (strncmp (*argv, "speed=", 6) == SAME) {

      			if ((baud = atoi (*argv+6)) == 0) {
				err ("Bad argument '%s'", argv[1]);
				continue;
			}
			switch (baud) {
			case 134:   ttyb.sg_ispeed = ttyb.sg_ospeed = B134; break;
			case 150:   ttyb.sg_ispeed = ttyb.sg_ospeed = B150; break;
			case 300:   ttyb.sg_ispeed = ttyb.sg_ospeed = B300; break;
			case 600:   ttyb.sg_ispeed = ttyb.sg_ospeed = B600; break;
			case 1200:  ttyb.sg_ispeed = ttyb.sg_ospeed = B1200; break;
			case 1800:  ttyb.sg_ispeed = ttyb.sg_ospeed = B1800; break;
			case 2400:  ttyb.sg_ispeed = ttyb.sg_ospeed = B2400; break;
			case 4800:  ttyb.sg_ispeed = ttyb.sg_ospeed = B4800; break;
			case 9600:  ttyb.sg_ispeed = ttyb.sg_ospeed = B9600; break;
			default:
				err("Unknown speed %d", baud);
				break;
			}
			continue;
		}
		if (strncmp (*argv, "parity=", 7) == SAME) {
			 parp = *argv+7;
			if (strncmp (parp, "odd", 3) == SAME) {
				ttyb.sg_flags = TANDEM|CRMOD|XTABS|ODDP;
			 } else if (strncmp (parp, "even", 4) == SAME) {
				ttyb.sg_flags = TANDEM|CRMOD|XTABS|EVENP;
#ifdef SEL
			 } else if (strncmp (parp, "mark", 4) == SAME) {
			 	ttyb.sg_flags = TANDEM|CRMOD|XTABS|MARKP;
			 } else if (strncmp (parp, "space", 4) == SAME) {
			 	ttyb.sg_flags = TANDEM|CRMOD|XTABS|SPACEP;
#endif
			} else {
				err("Unknown parity '%s'", parp);
				continue;
			}
		}	
	}
}

/*
 *		E R R ( )
 *	Called by banner() and its subfunctions, amongst others
 */
/* VARARGS */
err( fmt, a, b, c, d )
char *fmt;
{
	fprintf( stderr, "%s: ", arg0);
	fprintf( stderr, fmt, a, b, c, d );
	putc( '\n', stderr );
}


/*
 *	DORESTART ()
 *
 *  This function is called by a signal to restart a job.
 */

dorestart()
{
	putchar (CR);
	putchar (FF);	/* Send a cr and ff to clear current page */
	retmsg (RP_RESTART, "RESTART requested");
	exit( 0 );
}


/*
 *	DOFLUSH ()
 *
 *  This function is called by a signal to "flush" the job.
 */
doflush()
{
	putchar (CR);
	putchar (FF);	/* Send a cr and ff to clear current page */
	retmsg (RP_FLUSH, "Server received FLUSH request");
	exit( 0 );
}
